#include "utils.h"

#include <errno.h>

int json_get_array(json_object *json, const char *key, json_object **val)
{
	json_object *o;
	int r = -1;

	do {
		if (!fp_json_object_object_get_ex(json, key, &o)) break;
		if (!fp_json_object_is_type(o, json_type_array)) break;
		*val = o;
		r = fp_json_object_array_length(o);
	} while (0);

	return r;
}

int json_get_int32(json_object *json, const char *key, int32_t *val)
{
	json_object *o;
	int32_t v;
	int r = -1;

	do {
		if (!fp_json_object_object_get_ex(json, key, &o)) break;
		if (!fp_json_object_is_type(o, json_type_int)) break;
		errno = 0;
		v = fp_json_object_get_int(o);
		if (errno) break;
		*val = v;
		r = 0;
	} while (0);

	return r;
}

int json_get_int64(json_object *json, const char *key, int64_t *val)
{
	json_object *o;
	int64_t v;
	int r = -1;

	do {
		if (!fp_json_object_object_get_ex(json, key, &o)) break;
		if (!fp_json_object_is_type(o, json_type_int)) break;
		errno = 0;
		v = fp_json_object_get_int64(o);
		if (errno) break;
		*val = v;
		r = 0;
	} while (0);

	return r;
}

int json_get_double(json_object *json, const char *key, double *val)
{
	json_object *o;
	double v;
	int r = -1;

	do {
		if (!fp_json_object_object_get_ex(json, key, &o)) break;
		if (!fp_json_object_is_type(o, json_type_double)) break;
		errno = 0;
		v = fp_json_object_get_double(o);
		if (errno) break;
		*val = v;
		r = 0;
	} while (0);

	return r;
}

int json_get_string(json_object *json, const char *key, char **val)
{
	json_object *o;
	int r = -1;

	do {
		if (!fp_json_object_object_get_ex(json, key, &o)) break;
		if (!fp_json_object_is_type(o, json_type_string)) break;
		*val = (char *)fp_json_object_get_string(o);
		r = 0;
	} while (0);

	return r;
}

int json_get_boolean(json_object *json, const char *key, int *val)
{
	json_object *o;
	int r = -1;

	do {
		if (!fp_json_object_object_get_ex(json, key, &o)) break;
		if (!fp_json_object_is_type(o, json_type_boolean)) break;
		*val = fp_json_object_get_boolean(o);
		r = 0;
	} while (0);

	return r;
}
